/** ZCT/book.c--Created 091507 **/

/** Copyright 2008 Zach Wegner **/
/*
 * This file is part of ZCT.
 *
 * ZCT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ZCT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ZCT.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "zct.h"
#include "functions.h"
#include "globals.h"
#include "book.h"
#include "pgn.h"

FILE *book_file = NULL;

/**
book_update():
This is used to update an opening book from a given PGN file.
Created 091507; last modified 050408
**/
void book_update(char *pgn_file_name, char *book_file_name, int width, int depth, int win_percent)
{
	int buffer_size;
	int d;
	int game;
	int game_count;
	BOOK_PGN_INFO bpi;
	BOOK_POSITION *book_buffer;
	BOOK_POSITION *bp;

	/* Open the book file. */
	if (book_file != NULL)
		fclose(book_file);
	book_file = fopen(book_file_name, "wb+");
	if (book_file == NULL)
	{
		print("%s: could not open.\n", book_file_name);
		return;
	}
	/* Open the pgn file. */
	game_count = pgn_open(pgn_file_name);
	if (game_count == -1)
		return;
	if (game_count == 0)
	{
		print("%s: no games found.\n", pgn_file_name);
		return;
	}
	/* Allocate the buffer and start parsing. */
	buffer_size = (game_count * depth);
	print("Allocating book buffer of %i positions...\n", buffer_size);
	book_buffer = (BOOK_POSITION *)calloc(buffer_size, sizeof(BOOK_POSITION));
	if (book_buffer == NULL)
	{
		print("Failed to allocate book_buffer.\n");
		return;
	}
	bpi.book_buffer = book_buffer;
	bpi.buffer_size = buffer_size;
	bpi.depth = depth;
	/* Run through the PGN database and parse each game. Aren't PGN_FUNCs great? */
	for (game = 1; game <= game_count; game++)
		pgn_load(game, book_pgn_func, &bpi);
	/* Dump the book buffer into the book file. */
	print("Dumping positions to book... ");
	d = 0;
	for (bp = book_buffer; bp < book_buffer + buffer_size; bp++)
	{
		if (bp->hashkey != 0 && (bp->wins + bp->losses + bp->draws) >= width &&
			100 * (bp->wins + bp->draws / 2) / (bp->wins + bp->losses + bp->draws) >= win_percent)
		{
			fwrite(book_to_bytes(bp), sizeof(BOOK_POSITION), 1, book_file);
			d++;
		}
	}
	print("%i positions written.\n", d);
}

/**
book_load():
Loads an opening book given the file name.
Created 091507; last modified 091507
**/
void book_load(char *file_name)
{
	print("Trying opening book %s... ", file_name);
	if (book_file != NULL)
		fclose(book_file);
	book_file = fopen(file_name, "rb");
	if (book_file == NULL)
		print("%s: file not found.\n", file_name);
	else
		print("successful.\n");
}

/**
book_probe():
Probes the book for the current position. Suitable book moves are found, and a random move
is made from these.
Created 091507; last modified 050408
**/
BOOL book_probe(void)
{
	int book_position_count;
	int total_games;
	int random;
	BOOK_POSITION *bp;
	BOOK_POSITION book_positions[256];
	HASHKEY hashkey;
	ROOT_MOVE *root_move;
	unsigned char bytes[sizeof(BOOK_POSITION)];

	if (book_file == NULL)
		return FALSE;
	/* Probe the book to find all position with the same upper bits, i.e. from the same starting position. */
	book_position_count = 0;
	rewind(book_file);
	hashkey = board.hashkey & BOOK_UPPER_MASK;
	while (!feof(book_file))
	{
		if (fread(bytes, sizeof(bytes), 1, book_file))
		{
			bp = bytes_to_book(bytes);
			if ((bp->hashkey & BOOK_UPPER_MASK) == hashkey)
				book_positions[book_position_count++] = *bp;
		}
	}
	/* Now find the moves which are suitable. */
	generate_root_moves();
	total_games = 0;
	for (root_move = root_move_list; root_move < root_move_list + root_move_count; root_move++)
	{
		hashkey = board.hashkey & BOOK_UPPER_MASK;
		make_move(root_move->move);
		hashkey |= board.hashkey & BOOK_LOWER_MASK;
		unmake_move();
		for (bp = book_positions; bp < book_positions + book_position_count; bp++)
		{
			if (hashkey == bp->hashkey)
			{
				if (protocol == XBOARD)
				{
					if (total_games == 0)
						print("0 0 0 0 (");
					else
						print(" ");
					print("%M=%i", root_move->move, bp->wins + bp->losses + bp->draws);
				}
				else if (protocol != UCI)
					print("%M: games=%i winp=%.1f%%\n", root_move->move, bp->wins +
						bp->losses + bp->draws, (float)100 * (2 * bp->wins +
						bp->draws) / (bp->wins + bp->losses + bp->draws) / 2);
				total_games += bp->wins + bp->losses + bp->draws;
			}
		}
	}
	if (total_games <= 0)
		return FALSE;
	if (protocol == XBOARD)
		print(")\n");
	/* Now select the book move to be played. */
	random = (unsigned int)random_hashkey() % total_games;
	for (root_move = root_move_list; root_move < root_move_list + root_move_count; root_move++)
	{
		hashkey = board.hashkey & BOOK_UPPER_MASK;
		make_move(root_move->move);
		hashkey |= board.hashkey & BOOK_LOWER_MASK;
		unmake_move();
		for (bp = book_positions; bp < book_positions + book_position_count; bp++)
		{
			if (hashkey == bp->hashkey)
			{
				random -= bp->wins + bp->losses + bp->draws;
				if (random <= 0)
				{
					board.pv_stack[0][0] = root_move->move;
					return TRUE;
				}
			}
		}
	}
	return FALSE;
}

/**
book_to_bytes():
Converts an array of bytes into the internal BOOK_POSITION data structure. This is to allow
books created on any platform can be read on any platform.
Created 091507; last modified 101807
**/
BOOK_POSITION *bytes_to_book(unsigned char *bytes)
{
	static BOOK_POSITION bp;
	int x;

	bp.hashkey = (HASHKEY)0;
	for (x = 0; x < 8; x++)
		bp.hashkey |= (HASHKEY)*bytes++ << (x * 8);
	bp.wins = 0;
	for (x = 0; x < 2; x++)
		bp.wins |= *bytes++ << (x * 8);
	bp.losses = 0;
	for (x = 0; x < 2; x++)
		bp.losses |= *bytes++ << (x * 8);
	bp.draws = 0;
	for (x = 0; x < 2; x++)
		bp.draws |= *bytes++ << (x * 8);
	bp.flags = *bytes++;
	bp.learn = *bytes++;
	return &bp;
}

/**
book_to_bytes():
Converts the internal BOOK_POSITION data structure to an array of bytes. This is used to write
the book file portably on all platforms.
Created 091507; last modified 101807
**/
unsigned char *book_to_bytes(BOOK_POSITION *book_position)
{
	static unsigned char bytes[sizeof(BOOK_POSITION)];
	unsigned char *b;
	int x;

	b = bytes;
	for (x = 0; x < 8; x++)
		*b++ = (unsigned char)(book_position->hashkey >> (x * 8));
	for (x = 0; x < 2; x++)
		*b++ = (unsigned char)book_position->wins >> (x * 8);
	for (x = 0; x < 2; x++)
		*b++ = (unsigned char)book_position->losses >> (x * 8);
	for (x = 0; x < 2; x++)
		*b++ = (unsigned char)book_position->draws >> (x * 8);
	*b++ = (unsigned char)book_position->flags;
	*b++ = (unsigned char)book_position->learn;
	return bytes;
}

/**
book_pgn_func():
Called from within pgn_load, this function does all processing needed for a PGN game when creating a book.
It returns TRUE if we should stop parsing the PGN at this point.
Created 013008; last modified 050408
**/
BOOL book_pgn_func(void *pgn_arg, PGN_GAME *pgn_game)
{
	BOOK_PGN_INFO *bpi;
	BOOK_POSITION *bp;
	BOOK_POSITION *bp_start;
	BOOL drawn;
	COLOR winner;
	HASHKEY hashkey;
	
	bpi = (BOOK_PGN_INFO *)pgn_arg;
	/* Quickly determine whether we should stop parsing this game. */
	if (board.game_entry - board.game_stack > bpi->depth)
		return TRUE;

	/* Determine the game result. */
	winner = EMPTY;
	drawn = FALSE;
	if (!strncmp(pgn_game->tag.result, "1-0", 3))
		winner = WHITE;
	else if (!strncmp(pgn_game->tag.result, "0-1", 3))
		winner = BLACK;
	else if (!strncmp(pgn_game->tag.result, "1/2-1/2", 7))
		drawn = TRUE;
	/* No result, skip this game. */
	if (winner == EMPTY && drawn == FALSE)
		return TRUE;

	/* Update the book buffer for the move just made. */
	hashkey = (board.game_entry - 1)->hashkey & BOOK_UPPER_MASK;
	hashkey |= board.hashkey & BOOK_LOWER_MASK;
	/* Find an empty slot in the book buffer. */
	bp = bp_start = &bpi->book_buffer[hashkey % bpi->buffer_size];
	while (bp < bpi->book_buffer + bpi->buffer_size && bp->hashkey != 0 && bp->hashkey != hashkey)
		bp++;
	/* No slot found, go to beginning. */
	if (bp >= bpi->book_buffer + bpi->buffer_size)
	{
		bp = bpi->book_buffer;
		while (bp < bp_start && bp->hashkey != 0 && bp->hashkey != hashkey)
			bp++;
		/* Still no slot, buffer full. */
		if (bp >= bpi->book_buffer + bpi->buffer_size)
			return FALSE;
	}
	bp->hashkey = hashkey;
	if (winner == board.side_ntm)
		bp->wins++;
	else if (winner == board.side_tm)
		bp->losses++;
	else
		bp->draws++;
	return FALSE;
}
