/** ZCT/zct.c--Created 060205 **/

/** Copyright 2008 Zach Wegner **/
/*
 * This file is part of ZCT.
 *
 * ZCT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ZCT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ZCT.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "zct.h"
#include "functions.h"
#include "globals.h"
#include "eval.h"
#include "probe.h"
#include "smp.h"

int main(void)
{
	FILE *init_file;

	initialize_data();
	initialize_cmds();
	initialize_eval();
	initialize_attacks();
	initialize_board(NULL);
#ifdef SMP
	initialize_smp(MAX_CPUS);
#else
	process_count = 1;
#endif

#ifdef EGBB
	/* TODO: make this user-adjustable and SMP-safe. Note "safe" means it
		doesn't have to load all of it to RAM twice. */
	egbb_is_loaded = LoadEgbbLibrary("egbb/",EGBB_CACHE_SIZE);
#endif

	/* Read config file. */
	init_file = fopen("ZCT.ini", "rt");
	if (init_file != NULL)
	{
		source = TRUE;
		input_stream = init_file;
	}

	while (TRUE)
	{
		generate_root_moves();
		check_result(TRUE);
		if (protocol != ANALYSIS && protocol != ANALYSIS_X && protocol != DEBUG)
			update_clocks();
		if (ping)
		{
			if (protocol == XBOARD)
				print("pong %i\n", ping);
			else
				print("readyok\n");
			ping = 0;
		}
		if ((engine_state == DEBUGGING || protocol == ANALYSIS || board.side_tm == zct_side) && !game_over)
		{
			if (protocol == ANALYSIS || protocol == ANALYSIS_X)
				engine_state = ANALYZING;
			else if (engine_state != DEBUGGING && engine_state != INFINITE)
				engine_state = NORMAL;
			search_root();
			engine_state = IDLE;
			if (protocol != DEBUG)
			{
				if (board.pv_stack[0][0] != NULL_MOVE)
				{
					if (protocol == DEFAULT)
					{
						prompt();
						print("%M\n", board.pv_stack[0][0]);
					}
					else if (protocol == XBOARD)
						print("move %M\n", board.pv_stack[0][0]);
					else if (protocol == UCI)
					{
						print("bestmove %M", board.pv_stack[0][0]);
						if (board.pv_stack[0][1] != NULL_MOVE)
							print(" ponder %M", board.pv_stack[0][1]);
						print("\n");
						zct_side = EMPTY;
					}
					if (protocol != ANALYSIS && protocol != ANALYSIS_X && protocol != UCI)
					{
						make_move(board.pv_stack[0][0]);
						continue;
					}
				}
				else
					print("No move.\n");
			}
		}
		else if (hard && zct_side != EMPTY && board.game_entry > board.game_stack &&
			input_buffer[0] == '\0' && !game_over && protocol != UCI)
			ponder();
		prompt();
		/* Read input. */
		if (input_buffer == NULL || input_buffer[0] == '\0')
			read_line();
		else if (protocol != XBOARD && protocol != UCI)
			print("%s\n", input_buffer);
		if (command(input_buffer) == CMD_BAD && protocol != UCI)
			input_move(input_buffer, INPUT_USER_MOVE);
		/* Clear the input buffer. */
		input_buffer[0] = '\0';
	}
}

/**
prompt():
If the user is using console mode, this prompts for input depending on the engine mode.
Created 081906; last modified 012908
**/
void prompt(void)
{
	if (source)
		return;
	else if (protocol == DEFAULT)
		print("(zct)%i.%s ", board.move_number, board.side_tm == WHITE ? "" : "..");
	else if (protocol == DEBUG)
		print("(zct.debug)%i.%s ", board.move_number, board.side_tm == WHITE ? "" : "..");
	else if (protocol == ANALYSIS)
		print("(zct.analyze)%i.%s ", board.move_number, board.side_tm == WHITE ? "" : "..");
}

/**
generate_root_moves():
Sets up the root move list for either game play or searching.
Created 082606; last modified 110606
**/
void generate_root_moves(void)
{
	int next;
	static MOVE move_list[MAX_ROOT_MOVES];
	MOVE *last;

	last = generate_moves(move_list);
	check_legality(move_list, &last);
	for (next = 0; next < last - move_list; next++)
	{
		root_move_list[next].move = move_list[next];
		root_move_list[next].nodes = 0;
		root_move_list[next].score = 0;
		root_move_list[next].last_score = 0;
	}
	root_move_count = next;
}

/**
check_result():
Generates legal moves, checks if the game is over, and prints result if game is over.
Created 081906; last modified 050108
**/
BOOL check_result(BOOL print_result)
{
	MOVE move_list[256];
	MOVE *last;
	RESULT new_result;
   
	new_result = INCOMPLETE;
//	if (game_result != INCOMPLETE)
//		new_result = game_result;
//	else
	{
		last = generate_moves(move_list);
		check_legality(move_list, &last);
		if (last - move_list == 0)
		{
			if (in_check())
				new_result = (board.side_tm == WHITE ? BLACK_MATES : WHITE_MATES);
			else
				new_result = STALEMATE;
		}
		else if (board.fifty_count >= 100)
			new_result = FIFTY_DRAW;
		else if (is_repetition(3))
			new_result = THREE_REP_DRAW;
		else if (!can_mate(WHITE) && !can_mate(BLACK))
			new_result = MATERIAL_DRAW;
	}
	if (new_result != INCOMPLETE && print_result)
	{
		game_result = new_result;
		print("%s\n", result_str[game_result]);
		game_over = TRUE;
	}
	return (new_result != INCOMPLETE);
}

/**
bench():
Makes a standard measure of performance on a set of positions.
Created 013008; last modified 020208
**/
void bench(void)
{
	BOOL old_post;
	BITBOARD total_nodes;
	int time;
	struct bench_position
	{
		char *fen;
		int depth;
	} position[] =
	{
		{ "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq -", 13 },
		{ "2Q2bk1/5pp1/p3p2p/q1p1P2P/1pP5/1P2B1P1/P4PK1/8 w - -", 16 },
		{ "r1bqkb1r/pp1n1ppp/2p1pn2/8/2BP4/2N1PN2/PP3PPP/R1BQK2R b KQkq -", 12 },
		{ "rnbqk2r/pp2ppbp/3p1np1/8/3NP3/2N5/PPP1BPPP/R1BQK2R w KQkq -", 11 },
		{ "4q1kr/p6p/1prQPppB/4n3/4P3/2P5/PP2B2P/R5K1 w - -", 13 },
		{ "k2N4/1qpK1p2/1p6/1P4p1/1P4P1/8/8/8 w - -", 26 },
		{ "5k2/1p4pP/p7/1p1p4/8/2r3K1/6PP/8 w - -", 14 },
		{ NULL, 0 }
	};
	struct bench_position *current_position;

	input_buffer[0] = '\0';
	total_nodes = 0;
	time = get_time();
	use_book = FALSE;
	old_post = post;
	post = FALSE;
	engine_state = NORMAL;
	print("Running benchmark");
	/* Loop through the positions. */
	for (current_position = &position[0]; current_position->fen != NULL; current_position++)
	{
		print(".");
		initialize_board(current_position->fen);
		generate_root_moves();
		max_depth = current_position->depth * PLY;
		search_root();
		total_nodes += nodes + q_nodes;
		if (input_buffer[0])
			break;
	}
	time = get_time() - time;
	print("\n");
	print("Total nodes=%L\n", total_nodes);
	print("Total time=%T\n", time);
	print("Nodes per second=%L\n", total_nodes * 1000 / time);
	engine_state = IDLE;
	use_book = TRUE;
	post = old_post;
}
